import { NextFunction, Request, Response } from "express";
import prisma from "../../database/dbclient";
import { z } from "zod";

const attendanceSchema = z.object({
  company_id: z.coerce.number().int(),
  employee_id: z.coerce.number().int(),
  notes: z.coerce.number().int().nullable().optional(),
  shift_id: z.coerce.number().int().nullable().optional(),
  is_present: z.coerce.number().int().max(1),
});

export const createRandomAttendanceForEmployee = async (
  req: Request,
  res: Response,
  next: NextFunction
) => {
  const { employee_id, company_id, from_date, to_date } = req.body;
  // Dates
  const fd = new Date(from_date);
  const td = new Date(to_date);
  const fromDate = new Date(
    Date.UTC(fd.getFullYear(), fd.getMonth(), fd.getDate())
  );
  const toDate = new Date(
    Date.UTC(td.getFullYear(), td.getMonth(), td.getDate())
  );

  // Validate date range
  if (fromDate > toDate) {
    return res.status(400).json({ error: "Invalid date range" });
  }

  try {
    // Generate all dates in range
    console.log("POST > Generate Attendance > Date Data >", {
      fromDate,
      toDate,
    });

    const dates: Date[] = [];
    let currentDate = new Date(fromDate);
    while (currentDate <= toDate) {
      dates.push(new Date(currentDate));
      currentDate.setDate(currentDate.getDate() + 1);
    }

    // console.log(
    //   "POST > Generate Attendance > Date Data >",

    //   dates
    // );

    // console.log(
    //   "Test conflicting dates",
    //   await prisma.attendance_records.findMany({
    //     where: {
    //       company_id,
    //       employee_id,
    //       attendance_date: {
    //         lte: td,
    //         gte: fd,
    //       },
    //     },
    //     select: { record_id: true, attendance_date: true },
    //   })
    // );

    // Fetch existing attendance records & holidays
    const [conflictingDates, holidays, employeeShift] = await Promise.all([
      prisma.attendance_records.findMany({
        where: {
          company_id,
          employee_id,
          attendance_date: {
            lte: toDate,
            gte: fromDate,
          },
        },
        select: { record_id: true, attendance_date: true },
      }),

      prisma.holiday_list.findMany({
        where: {
          company_id,
          end_time: { lte: toDate },
          start_time: { gte: fromDate },
        },
        select: { start_time: true, end_time: true },
      }),

      // Get employee's shift
      prisma.duty_roaster.findFirst({
        where: { employee_id, company_id },
        select: {
          shift_db: {
            select: {
              shift_id: true,
              dayoff: true,
            },
          },
        },
      }),
    ]);

    console.log("POST > Generate Attendance > DB fetches done >", {
      conflictingDates,
      holidays,
      employeeShift,
    });

    if (!employeeShift) {
      return res
        .status(400)
        .json({ error: "No shift assigned to the employee" });
    }

    const { shift_id } = employeeShift.shift_db;
    const shiftDaysOff = employeeShift.shift_db.dayoff[0];

    // console.log("POST > Generate Attendance > Shift", shiftDaysOff);

    if (!shiftDaysOff) {
      return res
        .status(400)
        .json({ error: "No day-off settings found for the shift" });
    }

    // Identify weekends from shift settings
    const getAttendanceStatus = (date: Date) => {
      const dayName = date
        .toLocaleDateString("en-US", { weekday: "long" })
        .toLowerCase();
      if (shiftDaysOff[dayName as keyof typeof shiftDaysOff] === 0) return 2; // Weekend
      if (
        holidays.find((item) => {
          if (
            new Date(item.start_time).getTime() <= date.getTime() &&
            date.getTime() <= new Date(item.end_time).getTime()
          ) {
            // console.log("Here", item);
            return true;
          }
        })
      )
        return 3; // Holiday
      return Math.random() < 0.8 ? 1 : 0; // Random present (1) or absent (0)
    };

    // Process attendance updates (for existing records)
    const updateTasks = conflictingDates.map((cDate) => {
      return prisma.attendance_records.update({
        where: { record_id: cDate.record_id },
        data: { is_present: getAttendanceStatus(cDate.attendance_date) },
      });
    });

    // Create new attendance records
    const existingDateSet = new Set(
      conflictingDates.map((d) => d.attendance_date.toISOString())
    );
    const createTasks = dates
      .filter((date) => !existingDateSet.has(date.toISOString()))
      .map((date) => {
        return prisma.attendance_records.create({
          data: {
            company_id,
            employee_id,
            attendance_date: date,
            is_present: getAttendanceStatus(date),
          },
        });
      });

    // Execute all DB operations in a transaction
    const nr = await Promise.allSettled([...updateTasks, ...createTasks]);
    // const newRecords = await prisma.$transaction([
    //   ...updateTasks,
    //   ...createTasks,
    // ]);

    res.status(201).json({
      message: "Attendance data updated successfully",
      updated_records: nr
        .filter((item) => item.status === "fulfilled")
        .map((item) => item.value)
        .sort(
          (a, b) => a.attendance_date.getTime() - b.attendance_date.getTime()
        ),
    });
  } catch (error) {
    console.error(
      "POST > Generate Attendance > Failed to generate attendance >",
      error
    );
    next(error);
  }
};

export const addAttendance = async (
  req: Request,
  res: Response,
  next: NextFunction
) => {
  const parsed = attendanceSchema.safeParse(req.body);

  if (!parsed.success) {
    return res.status(400).json(parsed.error);
  }

  const { company_id, employee_id, is_present } = parsed.data;

  try {
    var todayDate = new Date();

    const dataExist = await prisma.attendance_records.findFirst({
      where: {
        AND: [{ employee_id: employee_id }, { attendance_date: todayDate }],
      },
    });

    if (dataExist == null) {
      return res.status(401).json({
        error: "Data Already Exist On the Same Date",
        data: dataExist,
      });
    }

    const addAttendanceObj = await prisma.attendance_records.create({
      data: {
        attendance_date: todayDate,
        is_present: is_present,
        employee_id: employee_id,
        company_id: company_id,
      },
    });

    res.status(200).json({ data: addAttendanceObj });
  } catch (error) {
    next(error);
  }
};

const bulkUpdateBodySchema = z.object({
  employee_id: z.number().positive(),
  company_id: z.number().positive(),
  value: z.number().default(0),
  from_date: z.string().datetime(),
  to_date: z.string().datetime(),
});

export const updateAttendance = async (
  req: Request,
  res: Response,
  next: NextFunction
) => {
  try {
    const parsed = bulkUpdateBodySchema.safeParse(req.body);

    if (parsed.error) {
      return res.status(400).json({
        message: "Invalid request body.",
        error: parsed.error,
      });
    }

    const { employee_id, company_id, from_date, to_date, value } = parsed.data;
    // Dates
    const fd = new Date(from_date);
    const td = new Date(to_date);
    const fromDate = new Date(
      Date.UTC(fd.getFullYear(), fd.getMonth(), fd.getDate())
    );
    const toDate = new Date(
      Date.UTC(td.getFullYear(), td.getMonth(), td.getDate())
    );

    // Date range validity check
    if (fromDate.getTime() > toDate.getTime()) {
      return res.status(400).json({
        message: "Invalid date range.",
        error: new Error(
          'Invalid date range. "From date" must preceed "To Date."'
        ),
      });
    }

    // Update many items
    const updatedRecords = await prisma.attendance_records.updateMany({
      where: {
        employee_id,
        company_id,
        attendance_date: {
          lte: toDate,
          gte: fromDate,
        },
      },
      data: {
        is_present: value,
      },
    });

    return res.status(200).json({
      message: "Update successful.",
      updatedRecords,
    });
  } catch (error) {
    console.error(
      "POST > Update multiple attendance > Failed to update >",
      error
    );
    next(error);
  }
};

export const generateAttendanceReport = async (
  req: Request,
  res: Response,
  next: NextFunction
) => {
  try {
    const {
      from_date,
      end_date,
      sort = "DESC",
      employee_id,
      page,
      limit,
    } = req.query;
    const pageNum = parseInt(page as string);
    const limitNum = parseInt(limit as string);

    const skip = (pageNum - 1) * limitNum;

    const { company_id } = req.params;

    const filters: any = {};

    if (company_id == null) {
      res.status(400).json({
        message: "Company Id Missing",
      });

      return;
    }

    filters.company_id = parseInt(company_id.toString());

    if (end_date != null) {
      const endDate = new Date(end_date.toString());
      endDate.setHours(23, 59, 59, 999); // due to lte as time is been saved

      filters.attendance_date = {
        ...filters.attendance_date,
        lte: endDate,
      };
    }

    if (from_date != null) {
      filters.attendance_date = {
        ...filters.attendance_date,
        gte: new Date(from_date.toString()),
      };
    }

    if (employee_id != null) {
      if (employee_id == "all") {
        filters.employee_id = { not: null };
      } else {
        filters.employee_id = parseInt(employee_id.toString());
      }
    }

    // Fetch data using Prisma with the dynamically built filters and sorting

    const reports = await prisma.attendance_records.findMany({
      where: filters,
      orderBy: {
        created_at: sort.toString().toLowerCase() === "desc" ? "desc" : "asc",
      },
      select: {
        shift_id: true,
        status_id: true,
        is_holiday: true,
        is_present: true,
        notes: true,
        attendance_date: true,
        employees: {
          select: {
            employee_id: true,
            designations: {
              select: {
                designation_name: true,
              },
            },
            image: true,
            user: {
              select: {
                last_name: true,
                first_name: true,
                middle_name: true,
                email: true,
              },
            },
          },
        },
      },

      take: limitNum,
      skip: skip,
    });

    const total_rows = await prisma.attendance_records.count({
      where: filters,
    });
    const total_page = Math.ceil(total_rows / limitNum);

    res.status(200).json({
      total_page,
      data_count: reports.length,
      next_page: pageNum < total_page ? pageNum + 1 : undefined,
      previous_page: pageNum > 1 ? pageNum - 1 : undefined,
      data: reports,
    });
  } catch (error) {
    next(error);
  }
};

export const generateAllAttendanceReport = async (
  req: Request,
  res: Response,
  next: NextFunction
) => {
  try {
    const { from_date, end_date, sort = "DESC", employee_id } = req.query;
    const { company_id } = req.params;

    const filters: any = {};

    if (company_id == null) {
      res.status(400).json({
        message: "Company Id Missing",
      });

      return;
    }

    if (end_date != null) {
      const endDate = new Date(end_date.toString());
      endDate.setHours(23, 59, 59, 999); // due to lte as time is been saved

      filters.attendance_date = {
        ...filters.attendance_date,
        lte: endDate,
      };
    }

    if (from_date != null) {
      filters.attendance_date = {
        ...filters.attendance_date,
        gte: new Date(from_date.toString()),
      };
    }

    filters.company_id = parseInt(company_id.toString());

    if (employee_id != null) {
      if (employee_id == "all") {
        filters.employee_id = { not: null };
      } else {
        filters.employee_id = parseInt(employee_id.toString());
      }
    }

    // Fetch data using Prisma with the dynamically built filters and sorting

    const reports = await prisma.attendance_records.findMany({
      where: filters,
      orderBy: {
        created_at: sort.toString().toLowerCase() === "desc" ? "desc" : "asc",
      },
      select: {
        shift_id: true,
        status_id: true,
        is_holiday: true,
        is_present: true,
        notes: true,
        attendance_date: true,
        employees: {
          select: {
            employee_id: true,
            designations: {
              select: {
                designation_name: true,
              },
            },
            image: true,
            user: {
              select: {
                last_name: true,
                first_name: true,
                middle_name: true,
                email: true,
              },
            },
          },
        },
      },
    });

    res.status(200).json(reports);
  } catch (error) {
    next(error);
  }
};

export const generateAbsentReport = async (
  req: Request,
  res: Response,
  next: NextFunction
) => {
  try {
    const {
      from_date,
      end_date,
      sort = "DESC",
      employee_id,
      page,
      limit,
    } = req.query;
    const pageNum = parseInt(page as string);
    const limitNum = parseInt(limit as string);

    const skip = (pageNum - 1) * limitNum;

    const { company_id } = req.params;

    const filters: any = {};

    if (company_id == null) {
      res.status(400).json({
        message: "Company Id Missing",
      });

      return;
    }

    if (end_date != null) {
      const endDate = new Date(end_date.toString());
      endDate.setHours(23, 59, 59, 999); // due to lte as time is been saved

      filters.attendance_date = {
        ...filters.attendance_date,
        lte: endDate,
      };
    }

    if (from_date != null) {
      filters.attendance_date = {
        ...filters.attendance_date,
        gte: new Date(from_date.toString()),
      };
    }
    filters.company_id = parseInt(company_id.toString());

    if (employee_id != null) {
      if (employee_id == "all") {
        filters.employee_id = { not: null };
      } else {
        filters.employee_id = parseInt(employee_id.toString());
      }
    }

    filters.is_present = 0;

    // Fetch data using Prisma with the dynamically built filters and sorting

    const reports = await prisma.attendance_records.findMany({
      where: filters,
      orderBy: {
        created_at: sort.toString().toLowerCase() === "desc" ? "desc" : "asc",
      },
      select: {
        shift_id: true,
        status_id: true,
        is_holiday: true,
        is_present: true,
        notes: true,
        attendance_date: true,
        employees: {
          select: {
            employee_id: true,
            designations: {
              select: {
                designation_name: true,
              },
            },
            image: true,
            user: {
              select: {
                last_name: true,
                first_name: true,
                middle_name: true,
                email: true,
              },
            },
          },
        },
      },

      take: limitNum,
      skip: skip,
    });

    const total_rows = await prisma.attendance_records.count({
      where: filters,
    });
    const total_page = Math.ceil(total_rows / limitNum);

    res.status(200).json({
      total_page,
      data_count: reports.length,
      next_page: pageNum < total_page ? pageNum + 1 : undefined,
      previous_page: pageNum > 1 ? pageNum - 1 : undefined,
      data: reports,
    });
  } catch (error) {
    next(error);
  }
};

export const generateAllAbsentReport = async (
  req: Request,
  res: Response,
  next: NextFunction
) => {
  try {
    const { from_date, end_date, sort = "DESC", employee_id } = req.query;

    const { company_id } = req.params;

    const filters: any = {};

    if (company_id == null) {
      res.status(400).json({
        message: "Company Id Missing",
      });

      return;
    }

    filters.company_id = parseInt(company_id.toString());

    if (end_date != null) {
      const endDate = new Date(end_date.toString());
      endDate.setHours(23, 59, 59, 999); // due to lte as time is been saved

      filters.attendance_date = {
        ...filters.attendance_date,
        lte: endDate,
      };
    }

    if (from_date != null) {
      filters.attendance_date = {
        ...filters.attendance_date,
        gte: new Date(from_date.toString()),
      };
    }

    if (employee_id != null) {
      if (employee_id == "all") {
        filters.employee_id = { not: null };
      } else {
        filters.employee_id = parseInt(employee_id.toString());
      }
    }

    filters.is_present = 0;

    // Fetch data using Prisma with the dynamically built filters and sorting

    const reports = await prisma.attendance_records.findMany({
      where: filters,
      orderBy: {
        created_at: sort.toString().toLowerCase() === "desc" ? "desc" : "asc",
      },
      select: {
        shift_id: true,
        status_id: true,
        is_holiday: true,
        is_present: true,
        notes: true,
        attendance_date: true,
        employees: {
          select: {
            employee_id: true,
            designations: {
              select: {
                designation_name: true,
              },
            },
            image: true,
            user: {
              select: {
                last_name: true,
                first_name: true,
                middle_name: true,
                email: true,
              },
            },
          },
        },
      },
    });

    res.status(200).json(reports);
  } catch (error) {
    next(error);
  }
};

//#region Att Cnfrm

export const checkTodaysAttendance = async (
  req: Request,
  res: Response,
  next: NextFunction
) => {
  const user_id = req.user_id;

  try {
    const today = new Date();
    const todayUTC = Date.UTC(
      today.getFullYear(),
      today.getMonth(),
      today.getDate()
    );
    const [attendanceData, empData] = await Promise.all([
      prisma.attendance_records.findFirst({
        where: {
          AND: {
            employees: {
              user_id,
            },
            attendance_date: {
              lte: new Date(todayUTC + 86400000 - 1), // Less than or equal to the last millisecond of the day
              gte: new Date(todayUTC), // Greater than or equal to the first millisecond of the day
            },
          },
        },
      }),
      prisma.employees.findUnique({
        where: {
          user_id,
        },
        select: {
          employee_id: true,
          company_id: true,
        },
      }),
    ]);

    if (!attendanceData)
      return res.status(200).json({ ...empData, is_present: 0 });
    else return res.status(200).json(attendanceData);
  } catch (error) {
    console.error(
      "GET > Check today's attendance of an employee > Faile to get attendance data >",
      error
    );
    next(error);
  }
};

const checkInReqBodSchema = z.object({
  company_id: z.number().positive(),
  employee_id: z.number().positive(),
});

export const checkInToday = async (
  req: Request,
  res: Response,
  next: NextFunction
) => {
  const user_id = req.user_id;

  try {
    const parsed = checkInReqBodSchema.safeParse(req.body);

    if (parsed.error) {
      return res.status(400).json({
        message: "Invalid request body.",
        error: parsed.error,
      });
    }

    const today = new Date();
    const todayUTC = Date.UTC(
      today.getFullYear(),
      today.getMonth(),
      today.getDate()
    );

    const attendanceData = await prisma.attendance_records.findFirst({
      where: {
        AND: {
          employees: {
            user_id,
          },
          attendance_date: {
            lte: new Date(todayUTC + 86400000 - 1), // Less than or equal to the last millisecond of the day
            gte: new Date(todayUTC), // Greater than or equal to the first millisecond of the day
          },
        },
      },
    });

    var newAttendanceData = undefined;

    if (!attendanceData) {
      newAttendanceData = await prisma.attendance_records.create({
        data: {
          company_id: parsed.data.company_id,
          employee_id: parsed.data.employee_id,
          attendance_date: new Date(todayUTC),
          check_in_time: new Date(),
          created_at: new Date(),
          is_present: 1,
        },
      });
    } else {
      newAttendanceData = await prisma.attendance_records.update({
        where: {
          record_id: attendanceData.record_id,
        },
        data: {
          is_present: 1,
        },
      });
    }

    return res.status(200).json(newAttendanceData);
  } catch (error) {
    console.error(
      "POST > Check In today's attendance of an employee > Faile to update attendance data >",
      error
    );
    next(error);
  }
};

//#endregion
